LLIST_init		PROC USES esi lpLList:DWORD,dwIniSize:DWORD
;================================================================
; lpLINKED LLIST_init(dwIniSize) - initialize a new linked list
;
; Paramters
;	lpLList		pointer to allocated space
;	dwIniSize	size of heap at the beginning
;
; Returns
;	none
;
; 1. Create private heap
; 2. Initialize list object
; 2. Create head and tail guardians
;
;================================================================
@1:		; Allocate space for list object
			;invoke GlobalAlloc,GPTR,sizeof LINKED_LIST
			;or	eax,eax
			;jz @11
		;==============
			mov	esi,lpLList
			ASSUME esi:PTR LINKED_LIST
		; create private heap
			invoke HeapCreate,HEAP_NO_SERIALIZE,dwIniSize,0		; allocate space for the members
			or	eax,eax
			jnz @2
		;==============
	@11:	; error occured
			;invoke GlobalFree,esi					; HeapCreate failed, drop llist object
			;xor	eax,eax						; return FALSE
			;jmp @End
		;==============
@2:		; Initialize linked list object
			xor	edx,edx
			mov	[esi].hHeap,eax					; handle of the private heap
			mov	[esi].dwLength,edx				; list is empty
			mov	[esi].lpHead,edx				; no head member
			mov	[esi].lpTail,edx				; no tail member
@3:		; Create head and tail sentinels
			invoke HeapAlloc,[esi].hHeap,HEAP_NO_SERIALIZE or HEAP_ZERO_MEMORY,sizeof LINKED_MEMBER*2
;PrintHex eax,'sentinels'
			or	eax,eax
			jz @End
			;==============
			;invoke LLIST_destroy,esi
			;xor	eax,eax
			;jmp @End
@@:		; Adjust guardians' links
			lea	ecx,[eax+sizeof LINKED_MEMBER]			; eax: head guardian; ecx: tail guardian
			mov	(LINKED_MEMBER PTR [eax]).lpPrevLink,NULL
			mov	(LINKED_MEMBER PTR [ecx]).lpNextLink,NULL
			mov	(LINKED_MEMBER PTR [ecx]).lpPrevLink,eax
			mov	(LINKED_MEMBER PTR [eax]).lpNextLink,ecx
			mov	[esi].lpHead,eax
			mov	[esi].lpTail,ecx
		; return lpLINKED
			mov	eax,esi
@End:
			ASSUME esi:NOTHING
			ret
LLIST_init		ENDP

LLIST_create		PROC USES esi dwIniSize:DWORD
;================================================================
; lpLINKED LLIST_create(dwIniSize) - create a new linked list
;
; Paramters
;	dwIniSize	size of heap at the beginning
;
; Returns
;	lpLList		pointer to LINKED_LIST object
;
; 1. Allocate space for list object, create private heap
; 2. Initialize list object
;
;================================================================
@1:		; Allocate space for list object
			invoke GlobalAlloc,GPTR,sizeof LINKED_LIST
			or	eax,eax
			jz @End
		;==============
			invoke LLIST_init,eax,dwIniSize
@End:
			ASSUME esi:NOTHING
			ret
LLIST_create		ENDP

LLIST_deinit		PROC USES esi lpLList:DWORD
;================================================================
; lpLINKED LLIST_deinit(lpLList) - destroy a linked list object
;
; Paramters
;	lpLList		pointer to LINKED_LIST object
;
; Returns
;	none
;
; 1. Destroy list object's private heap
; 2. Free up list object's space
;
;================================================================
			mov	esi,lpLList
			ASSUME esi:PTR LINKED_LIST
@1:		; Destroy linked list object's heap
			invoke HeapDestroy,[esi].hHeap

@2:		; Destroy linked list object
			;invoke GlobalFree,esi
@End:
			ASSUME esi:NOTHING
			ret
LLIST_deinit		ENDP

LLIST_destroy		PROC lpLList:DWORD
;================================================================
; void LLIST_destroy(lpLList) - destroy a linked list object
;
; Paramters
;	lpLList		pointer to LINKED_LIST object
;
; Returns
;	none
;
;================================================================
			invoke LLIST_deinit,lpLList
			invoke GlobalFree,lpLList
			ret
LLIST_destroy		ENDP

LLIST_insert		PROC USES esi lpLList:DWORD,dwIndex:DWORD,dwSize:DWORD
;================================================================
; lpMember LLIST_insert(lpLList,dwIndex,dwSize) - insert a linked list member
;
; Paramters
;	lpLList		pointer to LINKED_LIST object
;	dwIndex		index of new member
;			if zero, linked to the head
;			if -1, linked to the end
;	dwSize		size of new member in bytes
; Returns
;	lpMember	pointer to new list member object's user data
;			NULL when error
;
; 1. Determine insertion point
; 2. Allocate new member
; 3. Adjust links
;
;================================================================
LOCAL lpMember:DWORD
			mov	esi,lpLList
			ASSUME esi:PTR LINKED_LIST
@1:		; Determine insertion point
			invoke LLIST_retrieve,esi,dwIndex,NULL
			or	eax,eax
			jnz @F							; index not found
			mov	eax,[esi].lpTail				; insert as last member
		@@:	;==============
			mov	ecx,sizeof LINKED_MEMBER
			mov	lpMember,eax
			add	ecx,dwSize
@2:		; Allocate new member
			invoke HeapAlloc,[esi].hHeap,HEAP_ZERO_MEMORY or HEAP_NO_SERIALIZE,ecx
			or	eax,eax
			jz @End
@3:			;==============
			add	[esi].dwLength,1
		; Adjust links
			mov	ecx,dwSize
			mov	esi,lpMember
			ASSUME eax:PTR LINKED_MEMBER				; eax: m1
			ASSUME esi:PTR LINKED_MEMBER				; esi: m2
			mov	[eax].dwSize,ecx
			mov	[eax].lpNextLink,esi				; m1.next = m2
			mov	ecx,[esi].lpPrevLink				; ecx: m0
			mov	[esi].lpPrevLink,eax				; m2.prev = m1
			mov	[eax].lpPrevLink,ecx				; m1.prev = m0
			mov	(LINKED_MEMBER PTR [ecx]).lpNextLink,eax	; m0.next = m1
;PrintHex eax,'member'
@End:
			ASSUME esi:NOTHING
			ASSUME eax:NOTHING
			ret

LLIST_insert		ENDP

LLIST_remove		PROC USES esi lpLList:DWORD,dwIndex:DWORD,lpMember:DWORD
;================================================================
; void LLIST_remove(lpLList,dwIndex,lpMember) - remove a linked list member
;
; Paramters
;	lpLList		pointer to LINKED_LIST object
;	dwIndex		index of member
;	lpMember	pointer of member
;
; Returns
;	none
;
;================================================================
			mov	esi,lpLList
			ASSUME esi:PTR LINKED_LIST
		; Check member
			cmp	lpMember,NULL
			jnz @1
			invoke LLIST_retrieve,esi,dwIndex,lpMember
			or	eax,eax						; eax: m1
			jz @End
			jmp @F
		@1:
			mov	eax,lpMember
		@@:
			sub	[esi].dwLength,1
		; Unlink member
			mov	edx,(LINKED_MEMBER PTR [eax]).lpNextLink	; edx: m2	m1->m2
			mov	ecx,(LINKED_MEMBER PTR [eax]).lpPrevLink	; ecx: m0	m0<-m1
			mov	(LINKED_MEMBER PTR [ecx]).lpNextLink,edx	; m0->m2
			mov	(LINKED_MEMBER PTR [edx]).lpPrevLink,ecx	; m0<-m2
		; Destroy member
			invoke HeapFree,[esi].hHeap,HEAP_NO_SERIALIZE,eax	
@End:
			ret
LLIST_remove		ENDP

LLIST_retrieve		PROC USES ebx esi lpLList:DWORD,dwIndex:DWORD,lpMember:DWORD
;================================================================
; mixed LLIST_retrieve(lpLList,dwIndex) - look for a linked list member
;
; Paramters
;	lpLList		pointer to LINKED_LIST object
;	dwIndex		index of requested member
;	lpMember	pointer of requested member
;
; Returns
;	lpMember	pointer to list member object if lpMember argument was NULL
;	dwIndex		index to list member object if lpMember argument was supplied
;	NULL		if not found
;
;================================================================
			mov	esi,lpLList
			ASSUME esi:PTR LINKED_LIST
			mov	ebx,lpMember
			mov	ecx,dwIndex
			or	ebx,ebx
			jz @1
		; ================
		; Search by the address
			mov	edx,[esi].lpHead
			xor	eax,eax
ALIGN 4
@@:
			mov	ecx,(LINKED_MEMBER PTR [edx]).lpNextLink
			add	eax,1
			cmp	ecx,ebx
			jz @End
			mov	edx,ecx
			or	ecx,ecx
			jnz @B
			xor	eax,eax
			jmp @End
		; ================
@1:		; Search by the index
			mov	edx,[esi].dwLength
			xor	eax,eax
			cmp	ecx,edx
			jnc @End
		; ================
			add	ecx,1
			mov	eax,[esi].lpHead
ALIGN 4
@@:
			mov	eax,(LINKED_MEMBER PTR [eax]).lpNextLink
			or	eax,eax
			jz @End
			sub	ecx,1
			jnz @B
		; ================
@End:
			ret
			ASSUME esi:NOTHING
LLIST_retrieve		ENDP



LARRAY_init		PROC USES esi lpLArray:DWORD,dwElementSize:DWORD,dwSubArrayLen:DWORD
;================================================================
; lpLArray LARRAY_init(lpLArray,deElementSize,dwSubArrayLen) - create a new linked array
;
; Paramters
;	lpLArray	pointer to allocated space
;	deElementSize	size of one stored element in bytes
;	dwSubArrayLen	count of elements in one subarray
;
; Returns
;	none
;
;================================================================
			mov	esi,lpLArray
			ASSUME esi:PTR LINKED_ARRAY
		; create linked list of subarrays
			invoke LLIST_init,esi,0
			or	eax,eax
			jz @End
			xor	edx,edx
			mov	eax,dwElementSize
		; Align elementSize on DWORD
			push	eax
			add	eax,3
			and	eax,-4
			mov	[esi].dwElementSize,eax
		; Calculate size of subarray
			mov	ecx,dwSubArrayLen
			push	eax
			mul	ecx						; calculate subarray size in bytes
			add	eax,sizeof LINKED_SUBARRAY
			mov	[esi].dwSubArraySize,eax
		; Initialize array object
			mov	edx,1
			mov	[esi].dwSubArrayLen,ecx
			xor	eax,eax
			mov	[esi].dwFreeSlots,edx
			mov	[esi].dwElementCount,eax
			mov	eax,esi
@End:
			ret
			ASSUME esi:NOTHING
LARRAY_init		ENDP

LARRAY_create		PROC dwElementSize:DWORD,dwSubArrayLen:DWORD
;================================================================
; lpLArray LARRAY_create(dwElementSize,dwSubArrayLength) - create a new linked array
;
; Paramters
;	dwElementSize	size of one stored element in bytes
;	dwSubArrayLen	count of elements in one subarray
;
; Returns
;	lpLArray	pointer to LINKED_ARRAY object
;
;================================================================

			invoke GlobalAlloc,GPTR,sizeof LINKED_ARRAY
			or	eax,eax
			jz @End
			invoke LARRAY_init,eax,dwElementSize,dwSubArrayLen
@End:
			ret
LARRAY_create		ENDP

;LARRAY_destroy		PROC lpLArray:DWORD
;================================================================
; void LARRAY_destroy(lpLArray) - destroy a linked array object
;
; Paramters
;	lpLArray	pointer to LINKED_ARRAY object
;
; Returns
;	none
;
;================================================================
;			invoke LLIST_destroy,lpLArray
;
;			ret
;LARRAY_destroy		ENDP

LARRAY_insert		PROC USES esi lpLArray:DWORD
;================================================================
; lpVoid LARRAY_insert(lpLArray) - get next free linked array entry
;
; Paramters
;	lpLArray	pointer to LINKED_ARRAY object
; Returns
;	lpVoid		pointer to new array element
;			NULL when error
;
;================================================================
			mov	esi,lpLArray
			ASSUME esi:PTR LINKED_ARRAY
;mov	ecx,[esi].dwFreeSlots
;PrintHex ecx,'free'
			sub	[esi].dwFreeSlots,1
			jz @2
@1:	; Get next free slot
			mov	eax,[esi].dwSubArrayLen
			sub	eax,[esi].dwFreeSlots
			mul	[esi].dwElementSize
;PrintHex eax,'offset'
			mov	edx,[esi].lpTail
			add	eax,(LINKED_MEMBER PTR [edx]).lpPrevLink
			jmp @3
@2:	; Add new subarray
			push	[esi].dwSubArrayLen
			pop	[esi].dwFreeSlots
			invoke LLIST_insert,esi,-1,[esi].dwSubArraySize
;PrintHex eax,'subarray'
			or	eax,eax
			jz @End
@3:	; Return address
			add	[esi].dwElementCount,1
			add	eax,sizeof LINKED_SUBARRAY
;PrintHex eax,'element'
@End:
			ret
			ASSUME esi:NOTHING
LARRAY_insert		ENDP

LARRAY_remove		PROC USES edi esi lpLArray:DWORD,dwIndex:DWORD,lpElement:DWORD
;================================================================
; dwMoved LARRAY_remove(lpLArray,dwIndex,lpElement) - remove a linked array element
;
; Paramters
;	lpLArray	pointer to LINKED_ARRAY object
;	dwIndex		index of element
;	lpElement	pointer of requested member
;
; Returns
;	dwMoved		index of moved element
;
; 1. Copy last element into slot of removed element
; 2. if dwFreeSlots = dwSubArrayLen remove last subarray
;   
;
;================================================================
			mov	esi,lpLArray
			ASSUME esi:PTR LINKED_ARRAY
@1: ; Copy last element into slot of removed element
			invoke LARRAY_retrieve,esi,dwIndex,lpElement
			or	eax,eax
			jz @End
			mov	edi,eax

		; Get the last element
			mov	edx,[esi].lpTail
			mov	eax,[esi].dwSubArrayLen
			mov	ecx,(LINKED_MEMBER PTR [edx]).lpPrevLink
;PrintHex eax,'address'
			sub	eax,[esi].dwFreeSlots
			add	ecx,sizeof LINKED_SUBARRAY
			mul	[esi].dwElementSize
			add	eax,ecx
;PrintHex eax,'address'

			cmp	eax,edi
			jz @2
			mov	ecx,[esi].dwElementSize
ALIGN 4
		@@:
			mov	edx,DWORD PTR [edi+ecx]
			mov	DWORD PTR [eax+ecx],edx
			sub	ecx,4
			jnc @B
			; ========
	@2:
			mov	eax,[esi].dwFreeSlots
			sub	[esi].dwElementCount,1
			add	eax,1
			cmp	eax,[esi].dwSubArrayLen
			jna @F
			; ========
		; Remove last subarray
			mov	edx,[esi].lpTail
			invoke LLIST_remove,esi,NULL,(LINKED_MEMBER PTR [edx]).lpPrevLink
			mov	eax,1
		@@:
			mov	[esi].dwFreeSlots,eax
;PrintHex eax,'dwFreeSlots-------'
@End:
			ret
LARRAY_remove		ENDP

LARRAY_retrieve		PROC USES ebx edi esi lpLArray:DWORD,dwIndex:DWORD,lpElement:DWORD
;================================================================
; lpVoid LARRAY_retrieve(lpLArray,dwIndex) - look for a linked array element
;
; Paramters
;	lpLArray	pointer to LINKED_ARRAY object
;	dwIndex		index of requested member
;	lpElement	pointer of requested member
;
; Returns
;	lpElement	pointer to array element object if lpElement argument was NULL
;	dwIndex		index to array element object if lpElement argument was supplied
;	NULL		if not found
;
;================================================================
			mov	ebx,lpElement
			mov	esi,lpLArray
			mov	ecx,dwIndex
			ASSUME esi:PTR LINKED_ARRAY
			or	ebx,ebx
			jz @2_
		; ================
		; Search by the address
			mov	edi,[esi].lpHead
			mov	ecx,1
			mov	edx,[esi].dwSubArraySize
			sub	edx,sizeof LINKED_SUBARRAY
			sub	ebx,sizeof LINKED_SUBARRAY
ALIGN 4
@1:		; Get the subarray
		; eax: lower boundary; edx: upper boundary of subarray
;PrintHex ebx,'address'
			mov	eax,(LINKED_SUBARRAY PTR [edi]).lpNextLink
			lea	edi,[eax+edx]
			cmp	ebx,eax
			jc @F
			cmp	ebx,edi
			jc @11
		@@:
			add	ecx,[esi].dwSubArrayLen
			mov	edi,eax
			cmp	eax,[esi].lpTail
			jnz @1
		; ================
		; Not found
			xor	eax,eax
			jmp @End
	@11:	; ================
		; Found, now get the index
			sub	ebx,eax		; offset in subarray
			xor	edx,edx
			mov	eax,ebx
			div	DWORD PTR [esi].dwElementSize
			add	eax,ecx
			jmp @End
@2_:
;PrintHex ecx,'index'
@2:		; ================
		; Search by the index
			sub	ecx,1
			xor	eax,eax
			cmp	ecx,[esi].dwElementCount
			jnc @End
			; ========
			mov	eax,[esi].lpHead
ALIGN 4
	@21:	; Get the subarray
			mov	edx,(LINKED_SUBARRAY PTR [eax]).lpNextLink
			cmp	ecx,[esi].dwSubArrayLen
			jc @22
			; ========
			; mov	edx,(LINKED_SUBARRAY PTR [eax]).lpNextLink
			sub	ecx,[esi].dwSubArrayLen
			mov	eax,edx
			jmp @21
	@22:	; Get the element
			push	edx
			mov	eax,[esi].dwElementSize
			mul	ecx
			pop	edx
			add	eax,edx
			add	eax,sizeof LINKED_SUBARRAY
@3:	; Return element

@End:
;PrintHex eax,'return'
			ASSUME esi:NOTHING
			ret
LARRAY_retrieve		ENDP